import ast
import warnings

from ahk.script import ScriptEngine
from ahk.utils import escape_sequence_replace
from ahk.keys import Key
from ahk.directives import InstallKeybdHook, InstallMouseHook


class Hotkey:
    def __init__(self, engine: ScriptEngine, hotkey: str, script: str):
        """

        :param engine: an :py:class:`~ahk.AHK` instance
        :param hotkey: The hotkey to use (AutoHotkey syntax)
        :param script: The script to execute when the hotkey is activated (AutoHotkey code as a string)
        """
        self.hotkey = hotkey
        self.script = script
        self.engine = engine

    @property
    def running(self):
        return hasattr(self, '_proc')

    def _start(self, script):
        try:
            proc = self.engine.run_script(script, blocking=False)
            yield proc
        finally:
            self._stop()

    def start(self):
        """
        Starts an AutoHotkey process with the hotkey script
        """
        if self.running:
            raise RuntimeError('Hotkey is already running')
        script = self.engine.render_template('hotkey.ahk', blocking=False, script=self.script, hotkey=self.hotkey)
        self._gen = self._start(script)
        proc = next(self._gen)
        self._proc = proc

    def _stop(self):
        if not self.running:
            return
        self._proc.terminate()
        del self._proc

    def stop(self):
        """
        Stops the process if it is running
        """
        if not self.running:
            raise RuntimeError('Hotkey is not running')
        try:
            next(self._gen)
        except StopIteration:
            pass
        finally:
            del self._gen


class KeyboardMixin(ScriptEngine):
    def hotkey(self, *args, **kwargs):
        """
        Convenience function for creating ``Hotkey`` instance using current engine.

        :param hotkey: The hotkey to use (AutoHotkey syntax)
        :param script: The script to execute when the hotkey is activated (AutoHotkey code as a string)
        :return: an :py:class:`~ahk.keyboard.Hotkey` instance
        """
        return Hotkey(engine=self, *args, **kwargs)

    def key_state(self, key_name, mode=None) -> bool:
        """
        Check the state of a key.

        https://autohotkey.com/docs/commands/GetKeyState.htm

        :param key_name: the name of the key (or virtual key code)
        :param mode: see AHK docs
        :return: True if pressed down, else False
        """
        script = self.render_template('keyboard/key_state.ahk', key_name=key_name, mode=mode, directives=(InstallMouseHook, InstallKeybdHook))
        result = ast.literal_eval(self.run_script(script))
        return bool(result)

    def key_wait(self, key_name, timeout: int=None, logical_state=False, released=False):
        """
        Wait for key to be pressed or released (default is pressed; specify ``released=True`` to wait for key release).

        https://autohotkey.com/docs/commands/KeyWait.htm

        :param key_name: The name of the key
        :param timeout: how long (in seconds) to wait for the key. If not specified, waits indefinitely
        :param logical_state: Check the logical state of the key, which is the state that the OS and the active window believe the key to be in (not necessarily the same as the physical state). This option is ignored for joystick buttons.
        :param released: Set to True to wait for the key to be released rather than pressed
        :return: None
        :raises TimeoutError: if the key was not pressed (or released, if specified) within timeout
        """
        options = ''
        if not released:
            options += 'D'
        if logical_state:
            options += 'L'
        if timeout:
            options += f'T{timeout}'
        script = self.render_template('keyboard/key_wait.ahk', key_name=key_name, options=options)
        result = self.run_script(script)
        if result == "1":
            raise TimeoutError(f'timed out waiting for {key_name}')

    def type(self, s, blocking=True):
        """
        Sends keystrokes using send_input, also escaping the string for use in AHK.

        :param s: the string to type
        :param blocking: if ``True``, waits until script finishes, else returns immediately.
        """
        s = escape_sequence_replace(s)
        self.send_input(s, blocking=blocking)

    def send(self, s, raw=False, delay=None, blocking=True):
        """
        https://autohotkey.com/docs/commands/Send.htm

        :param s:
        :param raw:
        :param delay:
        :param blocking: if ``True``, waits until script finishes, else returns immediately.
        :return:
        """
        script = self.render_template('keyboard/send.ahk', s=s, raw=raw, delay=delay, blocking=blocking)
        self.run_script(script, blocking=blocking)

    def send_raw(self, s, delay=None):
        """
        https://autohotkey.com/docs/commands/Send.htm

        :param s:
        :param delay:
        :return:
        """
        return self.send(s, raw=True, delay=delay)

    def send_input(self, s, blocking=True):
        """
        https://autohotkey.com/docs/commands/Send.htm

        :param s:
        :param blocking:
        :return:
        """
        if len(s) > 5000:
            warnings.warn('String length greater than allowed. Characters beyond 5000 may not be sent. '
                          'See https://autohotkey.com/docs/commands/Send.htm#SendInputDetail for details.')

        script = self.render_template('keyboard/send_input.ahk', s=s, blocking=blocking)
        self.run_script(script, blocking=blocking)

    def send_play(self, s):
        """
        https://autohotkey.com/docs/commands/Send.htm


        :param s:
        :return:
        """
        script = self.render_template('keyboard/send_play.ahk', s=s)
        self.run_script(script)

    def send_event(self, s, delay=None):
        """
        https://autohotkey.com/docs/commands/Send.htm

        :param s:
        :param delay:
        :return:
        """
        script = self.render_template('keyboard/send_event.ahk', s=s, delay=delay)
        self.run_script(script)

    def key_press(self, key, release=True, blocking=True):
        """
        Press and (optionally) release a single key

        :param key:
        :param release:
        :return:
        """

        self.key_down(key, blocking=blocking)
        if release:
            self.key_up(key, blocking=blocking)

    def key_release(self, key, blocking=True):
        """
        Release a key that is currently in pressed down state

        :param key:
        :return:
        """
        if isinstance(key, str):
            key = Key(key_name=key)
        return self.send_input(key.UP, blocking=blocking)

    def key_down(self, key, blocking=True):
        """
        Press down a key (without releasing it)

        :param key:
        :param blocking:
        :return:
        """
        if isinstance(key, str):
            key = Key(key_name=key)
        return self.send_input(key.DOWN, blocking=blocking)

    def key_up(self, key, blocking=True):
        """
        Alias for :meth:~`KeyboardMixin.key_release`
        """
        return self.key_release(key, blocking=blocking)



"""
The ahk.keys module contains some useful constants and classes for working with keys.
"""


class Key:
    is_modifier = False
    symbol = ''

    def __init__(self, key_name):
        self._key_name = key_name

    @property
    def name(self):
        return self._key_name

    @property
    def DOWN(self):
        return '{' + f'{self.name} down' + '}'

    @property
    def UP(self):
        return '{' + f'{self.name} up' + '}'

    def __str__(self):
        return '{' + self.name + '}'

    def __hash__(self):
        return hash(str(self))

    def __mul__(self, n):
        if not isinstance(n, int):
            raise TypeError(f"Unsupported operand type(s) for *: '{self.__class__.__name__}' and '{type(n)}'")
        return '{' + f'{self.name} {n}' + '}'

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __add__(self, s):
        return str(self) + s

    def __repr__(self):
        return f'{self.__class__.__name__}(key_name={self.name!r})'

    def __format__(self, format_spec):
        return str(self)


SYMBOLS = {
    'Win': '#',
    'LWin': '<#',
    'RWin': '>#',
    'Shift': '+',
    'LShift': '<+',
    'RShift': '>+',
    'Alt': '!',
    'LAlt': '<!',
    'RAlt': '>!',
    'Control': '^',
    'LControl': '<^',
    'RControl': '>^',

}


class KeyCombo:
    def __init__(self, *modifiers):
        self._s = None
        self.modifiers = list(modifiers)
        assert all([isinstance(key, KeyModifier) for key in self.modifiers]), 'Keys must be modifiers'

    def __str__(self):
        s = ''.join(mod.symbol for mod in self.modifiers)
        if self._s is not None:
            s += self._s
        return s

    def __add__(self, other):
        if self._s is not None:
            raise ValueError('Key combo is already terminated')
        if isinstance(other, KeyCombo):
            combo = KeyCombo(*[*self.modifiers, *other.modifiers])
            if other._s:
                combo = combo + other._s
            return combo
        if isinstance(other, KeyModifier):
            self.modifiers.append(other)
        elif isinstance(other, Key) or isinstance(other, str):
            self._s = str(other)
            return self
        else:
            raise TypeError(f"unsupported operand type(s) for +: '{self.__class__.__name__}' and '{type(other)}'")

    def __repr__(self):
        key_modifiers = ', '.join(repr(mod) for mod in self.modifiers)
        return f'{self.__class__.__name__}({key_modifiers}){f"+{self._s!r}" if self._s else f""}'


class KeyModifier(Key):
    is_modifier = True

    @property
    def symbol(self):
        return SYMBOLS.get(self.name, str(self))

    def __add__(self, other):
        if isinstance(other, KeyModifier):
            return KeyCombo(self, other)
        elif isinstance(other, KeyCombo):
            return other + self

        return self.symbol + str(other)


class KEYS:
    """
    KEYS constants
    REF: https://autohotkey.com/docs/KeyList.htm
    """
    CAPS_LOCK = Key('CapsLock')
    CapsLock = CAPS_LOCK
    SCROLL_LOCK = Key('ScrollLock')
    ScrollLock = SCROLL_LOCK
    SPACE = Key('Space')

    TAB = Key('Tab')
    Tab = TAB
    ENTER = Key('Enter')
    Enter = ENTER
    ESCAPE = Key('Escape')
    BACKSPACE = Key('Backspace')
    Backspace = BACKSPACE
    UP = Key('Up')
    Up = UP
    DOWN = Key('Down')
    Down = DOWN
    LEFT = Key('Left')
    Left = LEFT
    RIGHT = Key('Right')
    Right = RIGHT
    DELETE = Key('Delete')
    DEL = DELETE
    Delete = DELETE
    Del = DELETE

    WIN = KeyModifier('Win')
    Win = WIN
    LEFT_WIN = KeyModifier('LWin')
    LWin = LEFT_WIN
    RIGHT_WIN = KeyModifier('RWin')
    RWin = RIGHT_WIN
    CONTROL = KeyModifier('Control')
    Control = CONTROL
    CTRL = CONTROL
    Ctrl = CONTROL
    LEFT_CONTROL = KeyModifier('LControl')
    LCtrl = LEFT_CONTROL
    LControl = LEFT_CONTROL
    RIGHT_CONTROL = KeyModifier('RControl')
    RCtrl = RIGHT_CONTROL
    RControl = RIGHT_CONTROL
    ALT = KeyModifier('Alt')
    Alt = ALT
    LEFT_ALT = KeyModifier('LAlt')
    LAlt = LEFT_ALT
    RIGHT_ALT = KeyModifier('RAlt')
    RAlt = RIGHT_ALT
    SHIFT = KeyModifier('Shift')
    Shift = SHIFT
    LEFT_SHIFT = KeyModifier("LShift")
    LShift = LEFT_SHIFT
    RIGHT_SHIFT = KeyModifier('RShift')
    RShift = RIGHT_SHIFT
    NUMPAD_DOT = Key('NumpadDot')
    NumpadDot = NUMPAD_DOT
    NUMPAD_DEL = Key('NumpadDel')
    NumpadDel = NUMPAD_DEL
    NUM_LOCK = Key('NumLock')
    NumLock = NUM_LOCK
    NUMPAD_ADD = Key('NumpadAdd')
    NUMPAD_DIV = Key('NumpadDiv')
    NUMPAD_SUB = Key('NumpadSub')
    NUMPAD_MULT = Key('NumpadMult')
    NUMPAD_ENTER = Key('NumpadEnter')
    NumpadAdd = NUMPAD_ADD
    NumpadDiv = NUMPAD_DIV
    NumpadSub = NUMPAD_SUB
    NumpadMult = NUMPAD_MULT
    NumpadEnter = NUMPAD_ENTER


def _init_keys():
    '''put this in a function to avoid polluting global namespace'''
    for i in range(0, 10):
        #  set numpad keys
        key_name = f'Numpad{i}'
        key = Key(key_name)
        setattr(KEYS, key_name, key)
        setattr(KEYS, key_name.upper(), key)

    for i in range(1, 25):
        # set function keys
        key_name = f'F{i}'
        setattr(KEYS, key_name, Key(key_name))

    for i in range(1, 33):
        # set joystick keys
        key_name = f'Joy{i}'
        setattr(KEYS, key_name, Key(key_name))
        setattr(KEYS, key_name.upper(), Key(key_name))


_init_keys()

__all__ = [name for name in dir(KEYS) if not name.startswith('_')]


def __getattr__(name):
    return getattr(KEYS, name)



import logging

ESCAPE_SEQUENCE_MAP = {
    '\n': '`n',
    '\t': '`t',
    '\r': '`r',
    '\a': '`a',
    '\b': '`b',
    '\f': '`f',
    '\v': '`v',
    ',': '`,',
    '%': '`%',
    '`': '``',
    ';': '`;',
    ':': '`:',
    '!': '{!}',
    '^': '{^}',
    '+': '{+}',
    '{': '{{}',
    '}': '{}}',
    '#': '{#}'
}

_TRANSLATION_TABLE = str.maketrans(ESCAPE_SEQUENCE_MAP)

def make_logger(name):
    logger = logging.getLogger(name)
    handler = logging.NullHandler()
    formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def escape_sequence_replace(s):
    """
    Replace Python escape sequences with AHK equivalent escape sequences
    Additionally escapes some other characters for AHK escape sequences.
    Intended for use with AHK Send command functions.

    Note: This DOES NOT provide ANY assurances against accidental or malicious injection. Does NOT escape quotes.

    >>> escape_sequence_replace('Hello, World!')
    'Hello`, World{!}'
    """
    return s.translate(_TRANSLATION_TABLE)




from ahk.script import ScriptEngine
import os


class RegisteryMixin(ScriptEngine):

    def _render_template(self, template_name, *args, **kwargs):
        return self.render_template(
            os.path.join("registery", template_name),
        )

    def _run_template(self, template_name, *args, **kwargs):
        script = self._render_template(
            template_name,
            *args,
            **kwargs
        )
        return self.run_script(script)

    def read(self, key_name: str, value_name="") -> str:
        """Read registery

            Reference:
                https://www.autohotkey.com/docs/commands/RegRead.htm

            Arguments:
                key_name {str} -- RegEdit

            Keyword Arguments:
                value_name {str} -- TODO (default: {""})

            Returns:
                str -- Registery value
            """
        self._run_template(
            "reg_read.ahk",
            key_name=key_name,
            value_name=value_name
        )

    def delete(self, key_name: str, value_name="") -> None:
        """Delete registery

            Reference:
                https://www.autohotkey.com/docs/commands/RegDelete.htm

            Arguments:
                key_name {str} -- RegEdit

            Keyword Arguments:
                value_name {str} -- TODO (default: {""})
            """
        self._run_template(
            "reg_delete.ahk",
            key_name=key_name,
            value_name=value_name
        )

    def write(self, value_type: str, key_name: str, value_name="") -> None:
        """Write registery

            Reference:
                https://www.autohotkey.com/docs/commands/RegWrite.htm

            Arguments:
                value_type {str} -- RegEdit value
                key_name {str} -- RegEdit

            Keyword Arguments:
                value_name {str} -- TODO (default: {""})
            """
        self._run_template(
            "reg_write.ahk",
            value_type=value_type,
            key_name=key_name,
            value_name=value_name
        )

    def set_view(self, reg_view: int) -> None:
        """Set registery view

            Reference:
                https://www.autohotkey.com/docs/commands/SetRegView.htm

            Arguments:
                reg_view {str} -- Registery view
            """

        if reg_view not in [32, 64, "32", "64"]:
            raise ValueError("No valid bit, please use 32 or 64")

        self._run_template(
            "reg_set_view.ahk",
            reg_view=reg_view,
        )

    def loop(self, reg: str, key_name: str, mode=""):
        """Loop registery

        Reference:
            https://www.autohotkey.com/docs/commands/LoopReg.htm

        Arguments:
            reg {str} -- TODO
            key_name {str} -- TODO

        Keyword Arguments:
                mode {str} -- TODO (default: {""})
        """
        raise NotImplementedError

        self._run_template(
            "reg_loop.ahk",
            reg=reg,
            key_name=key_name,
            mode=mode
        )



from collections import deque
from ahk.mouse import MouseMixin
from ahk.window import Window, WindowMixin
from ahk.script import ScriptEngine
from ahk.screen import ScreenMixin
from ahk.keyboard import KeyboardMixin
from ahk.sound import SoundMixin
from ahk.registery import RegisteryMixin


class AHK(
    WindowMixin, MouseMixin, KeyboardMixin,
    ScreenMixin, SoundMixin, RegisteryMixin
):
    """
    Inherits its methods from the following classes:

    | :py:class:`ahk.window.WindowMixin`
    | :py:class:`ahk.mouse.MouseMixin`
    | :py:class:`ahk.keyboard.KeyboardMixin`
    | :py:class:`ahk.screen.ScreenMixin`
    | :py:class:`ahk.sound.SoundMixin`
    | :py:class:`ahk.registery.RegisteryMixin`
    """
    pass


class ActionChain(AHK):
    """
    Reusable action chain to execute various actions in order
    """
    def __init__(self, *args, **kwargs):
        self._actions = deque()
        super().__init__(*args, **kwargs)

    def run_script(self, *args, **kwargs):
        """
        override run_script to defer to queue
        """
        kwargs['decode'] = False  #
        self._actions.appendleft((args, kwargs))

    def perform(self):
        results = []
        while self._actions:
            args, kwargs = self._actions.pop()
            results.append(super().run_script(*args, **kwargs))
        return results

    def sleep(self, n):
        """
        :param n: how long (in seconds) to sleep
        :return:
        """
        n = n * 1000  # convert to milliseconds
        script = self.render_template('base.ahk', body=f'Sleep {n}', directives={'#Persistent', })
        self.run_script(script)



"""
The :py:mod:`~ahk.script` module, most essentially, houses the :py:class:`~ahk.ScriptEngine` class.

The :py:class:`~ahk.ScriptEngine` is responsible for rendering autohotkey code from jinja templates and executing that
code. This is the heart of how this package works. Every other major component either inherits from this class
or utilizes an instance of this class.

The current implementation of how autohotkey code is executed is by calling the autohotkey
executable with ``subprocess``.


"""
import os
import subprocess
import warnings
from shutil import which
from ahk.utils import make_logger
from ahk.directives import Persistent
from jinja2 import Environment, FileSystemLoader

logger = make_logger(__name__)


class ExecutableNotFoundError(EnvironmentError):
    pass


DEFAULT_EXECUTABLE_PATH = r"C:\Program Files\AutoHotkey\AutoHotkey.exe"
"""The deafult path to look for AutoHotkey, if not specified some other way"""


def _resolve_executable_path(executable_path: str = ''):
    if not executable_path:
        executable_path = os.environ.get('AHK_PATH') or which(
            'AutoHotkey.exe') or which('AutoHotkeyA32.exe')

    if not executable_path:
        if os.path.exists(DEFAULT_EXECUTABLE_PATH):
            executable_path = DEFAULT_EXECUTABLE_PATH

    if not executable_path:
        raise ExecutableNotFoundError(
            'Could not find AutoHotkey.exe on PATH. '
            'Provide the absolute path with the `executable_path` keyword argument '
            'or in the AHK_PATH environment variable.'
        )

    if not os.path.exists(executable_path):
        raise ExecutableNotFoundError(
            f"executable_path does not seems to exist: '{executable_path}' not found")

    if os.path.isdir(executable_path):
        raise ExecutableNotFoundError(
            f"The path {executable_path} appears to be a directory, but should be a file."
            " Please specify the *full path* to the autohotkey.exe executable file"
        )

    if not executable_path.endswith('.exe'):
        warnings.warn(
            'executable_path does not appear to have a .exe extension. This may be the result of a misconfiguration.'
        )

    return executable_path


class ScriptEngine(object):

    def __init__(self, executable_path: str = "", **kwargs):
        """
        This class is typically not used directly. AHK components inherit from this class
        and the arguments for this class should usually be passed in to :py:class:`~ahk.AHK`.

        :param executable_path: the path to the AHK executable.
            If not provided explicitly in this argument, the path to the AHK executable is resolved in the following order:

              * The ``AHK_PATH`` environment variable, if present
              * :py:data:`~ahk.script.DEFAULT_EXECUTABLE_PATH` if the file exists

            If environment variable not present, tries to look for 'AutoHotkey.exe' or 'AutoHotkeyA32.exe' with shutil.which

        :raises ExecutableNotFound: if AHK executable cannot be found or the specified file does not exist
        """
        self.executable_path = _resolve_executable_path(executable_path)

        templates_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
        self.env = Environment(loader=FileSystemLoader(templates_path),
                               autoescape=False, trim_blocks=True)

    def render_template(self, template_name, directives=None, blocking=True, **kwargs):
        """
        Renders a given jinja template and returns a string of script text

        :param template_name: the name of the jinja template to render
        :param directives: additional AHK directives to add to the resulting script
        :param blocking: whether the template should be rendered to block (use #Persistent directive)
        :param kwargs: keywords passed to template rendering
        :return: An AutoHotkey script as a string

        .. code-block:: python

            >>> from ahk import AHK
            >>> ahk = AHK()
            >>> ahk.render_template('keyboard/send_input.ahk', s='Hello')
            '#NoEnv\\n#Persistent\\n\\n\\nSendInput Hello\\n\\nExitApp\\n'
        """
        if directives is None:
            directives = set()
        else:
            directives = set(directives)
        if blocking:
            directives.add(Persistent)
        elif Persistent in directives:
            directives.remove(Persistent)

        kwargs['directives'] = directives
        template = self.env.get_template(template_name)
        return template.render(**kwargs)

    def _run_script(self, script_text, **kwargs):
        blocking = kwargs.pop('blocking', True)
        runargs = [self.executable_path, '/ErrorStdOut', '*']
        decode = kwargs.pop('decode', False)
        script_bytes = bytes(script_text, 'utf-8')
        if blocking:
            result = subprocess.run(runargs, input=script_bytes,
                                    stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kwargs)
            if decode:
                logger.debug('Stdout: %s', repr(result.stdout))
                logger.debug('Stderr: %s', repr(result.stderr))
                return result.stdout.decode()
            else:
                return result
        else:
            proc = subprocess.Popen(runargs, stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
            try:
                proc.communicate(script_bytes, timeout=0)
            except subprocess.TimeoutExpired:
                pass  # for now, this seems needed to avoid blocking and use stdin
            return proc

    def run_script(self, script_text: str, decode=True, blocking=True, **runkwargs):
        """
        Given an AutoHotkey script as a string, execute it

        :param script_text: a string containing AutoHotkey code
        :param decode: If ``True``, attempt to decode the stdout of the completed process.
            If ``False``, returns the completed process. Only has effect when ``blocking=True``
        :param blocking: If ``True``, script must finish before returning.
            If ``False``, function returns a ``subprocess.Popen`` object immediately without blocking
        :param runkwargs: keyword arguments passed to ``subprocess.Popen`` or ``subprocess.run``
        :return: | A string of the decoded stdout if ``blocking`` and ``decode`` are True.
                 | ``subprocess.CompletedProcess`` if ``blocking`` is True and ``decode`` is False.
                 | ``subprocess.Popen`` object if ``blocking`` is False.

        >>> from ahk import AHK
        >>> ahk = AHK()
        >>> ahk.run_script('FileAppend, Hello World, *')
        'Hello World'
        >>> ahk.run_script('FileAppend, Hello World, *', decode=False)
        CompletedProcess(args=['C:\\\\pathto\\\\AutoHotkey.exe', '/ErrorStdOut', '*'], returncode=0, stdout=b'Hello World', stderr=b'')
        >>> ahk.run_script('FileAppend, Hello World, *', blocking=False)
        <subprocess.Popen at 0x18a599cde10>
        """
        logger.debug('Running script text: %s', script_text)
        try:
            result = self._run_script(script_text, decode=decode, blocking=blocking, **runkwargs)
        except Exception as e:
            logger.fatal('Error running temp script: %s', e)
            raise
        return result



import ast
from contextlib import suppress

from ahk.script import ScriptEngine
from ahk.utils import escape_sequence_replace, make_logger

logger = make_logger(__name__)


class WindowNotFoundError(ValueError):
    pass


class Control:
    def __init__(self):
        raise NotImplementedError

    def click(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlClick.htm

        :return:
        """
        raise NotImplementedError

    def focus(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlFocus.htm

        :return:
        """
        raise NotImplementedError

    def get(self, key):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGet.htm

        :param key:
        :return:
        """
        raise NotImplementedError

    def has_focus(self):
        raise NotImplementedError

    @property
    def position(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGetPos.htm

        :return:
        """
        raise NotImplementedError

    @property
    def text(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGetText.htm

        :return:
        """
        raise NotImplementedError

    @text.setter
    def text(self, new_text):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlSetText.htm

        :param new_text:
        :return:
        """
        raise NotImplementedError

    def move(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlMove.htm

        :return:
        """
        raise NotImplementedError

    def send(self, raw=False):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlSend.htm

        :param raw:
        :return:
        """
        raise NotImplementedError


class Window(object):

    MINIMIZED = "-1"
    MAXIMIZED = "1"
    NON_MIN_NON_MAX = "0"

    _set_subcommands = {
        'always_on_top': 'AlwaysOnTop',
        'bottom': 'Bottom',
        'top': 'Top',
        'disable': 'Disable',
        'enable': 'Enable',
        'redraw': 'Redraw',
        'style': 'Style',
        'ex_style': 'ExStyle',
        'region': 'Region',
        'transparent': 'Transparent',
        'transcolor': 'TransColor'
    }

    _get_subcommands = {
        'id': 'ID',
        'id_last': 'IDLast',
        'pid': 'PID',
        'process_name': 'ProcessName',
        'process_path': 'ProcessPath',
        'process': 'ProcessPath',
        'count': 'count',
        'list': 'list',
        'min_max': "MinMax",
        'controls': 'ControlList',
        'controls_hwnd': 'ControlListHwnd',
        'transparent': 'Transparent',
        'trans_color': 'TransColor',
        'style': 'Style',   # This will probably get a property later
        'ex_style': 'ExStyle',  # This will probably get a property later
    }

    #  add reverse lookups
    _set_subcommands.update({value: value for value in _set_subcommands.values()})
    _get_subcommands.update({value: value for value in _get_subcommands.values()})

    def __init__(self, engine: ScriptEngine, ahk_id: str, encoding=None):
        self.engine = engine  # should this be a weakref instead?
        self.id = ahk_id
        self.encoding = encoding

    @classmethod
    def from_mouse_position(cls, engine: ScriptEngine, **kwargs):
        script = engine.render_template('window/from_mouse.ahk')
        ahk_id = engine.run_script(script)
        return cls(engine=engine, ahk_id=ahk_id, **kwargs)

    @classmethod
    def from_pid(cls, engine: ScriptEngine, pid, **kwargs):
        script = engine.render_template('window/get.ahk',
                                        subcommand="ID",
                                        title=f'ahk_pid {pid}')
        ahk_id = engine.run_script(script)
        return cls(engine=engine, ahk_id=ahk_id, **kwargs)

    def __getattr__(self, attr):
        if attr.lower() in self._get_subcommands:
            return self.get(attr)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def get(self, subcommand):
        sub = self._get_subcommands.get(subcommand)
        if not sub:
            raise ValueError(f'No such subcommand {subcommand}')

        script = self._render_template(
            'window/get.ahk',
            subcommand=sub,
            title=f"ahk_id {self.id}",
        )

        return self.engine.run_script(script)

    def __repr__(self):
        return f'<ahk.window.Window ahk_id={self.id}>'

    def set(self, subcommand, value):
        sub = self._set_subcommands.get(subcommand)
        if not sub:
            raise ValueError(f'No such subcommand {subcommand}')

        script = self._render_template(
            'window/win_set.ahk',
            subcommand=subcommand,
            value=value,
            title=f"ahk_id {self.id}"
        )
        return self.engine.run_script(script)

    def _get_pos(self):
        script = self._render_template(
            'window/win_position.ahk',
            title=f"ahk_id {self.id}"
        )
        resp = self.engine.run_script(script)
        try:
            value = ast.literal_eval(resp)
            return value
        except SyntaxError:
            raise WindowNotFoundError('No window found')

    @property
    def rect(self):
        return self._get_pos()

    @rect.setter
    def rect(self, new_position):
        x, y, width, height = new_position
        self.move(x=x, y=y, width=width, height=height)

    @property
    def position(self):
        x, y, _, _ = self._get_pos()

        return x, y

    @position.setter
    def position(self, new_position):
        x, y = new_position
        self.move(x, y)

    @property
    def width(self):
        _, _, width, _ = self._get_pos()
        return width

    @width.setter
    def width(self, new_width):
        self.move(width=new_width)

    @property
    def height(self):
        _, _, _, height = self._get_pos()
        return height

    @height.setter
    def height(self, new_height):
        self.move(height=new_height)

    def _base_property(self, command):
        script = self._render_template(
            "window/base_check.ahk",
            command=command,
            title=f"ahk_id {self.id}"
        )
        resp = self.engine.run_script(script)
        return bool(ast.literal_eval(resp))

    @property
    def active(self):
        return self._base_property(command="WinActive")

    @property
    def exist(self):
        return self._base_property(command="WinExist")

    def _base_get_method(self, command):
        script = self._render_template(
            "window/base_get_command.ahk",
            command=command,
            title=f"ahk_id {self.id}"
        )
        result = self.engine.run_script(script, decode=False)
        if self.encoding:
            return result.stdout.decode(encoding=self.encoding)
        return result.stdout

    @property
    def title(self):
        return self._base_get_method("WinGetTitle")

    @title.setter
    def title(self, value):
        script = self._render_template(
            "window/win_set_title.ahk",
            title=f"ahk_id {self.id}",
            new_title=value
        )
        return self.engine.run_script(script)

    @property
    def class_name(self):
        return self._base_get_method("WinGetClass")

    @property
    def text(self):
        return self._base_get_method("WinGetText")

    @property
    def minimized(self):
        return self.get("MinMax") == self.MINIMIZED

    @property
    def maximized(self):
        return self.get("MinMax") == self.MAXIMIZED

    @property
    def non_max_non_min(self):
        return self.get("MinMax") == self.NON_MIN_NON_MAX

    @property
    def transparent(self) -> int:
        result = self.get("Transparent")
        if result:
            return int(result)
        else:
            return 255

    @transparent.setter
    def transparent(self, value):
        if isinstance(value, int) and 0 <= value <= 255:
            self.set("Transparent", value)
        else:
            raise ValueError(
                f'"{value}" not a valid option. Please use [0, 255] integer')

    @property
    def always_on_top(self) -> bool:
        script = self._render_template(
            'window/win_is_always_on_top.ahk',
            title=f"ahk_id {self.id}"
        )
        resp = self.engine.run_script(script)
        return bool(ast.literal_eval(resp))

    @always_on_top.setter
    def always_on_top(self, value):
        if value in ('on', 'On', True, 1):
            self.set('AlwaysOnTop', 'On')
        elif value in ('off', 'Off', False, 0):
            self.set('AlwaysOnTop', 'Off')
        elif value in ('toggle', 'Toggle', -1):
            self.set('AlwaysOnTop', 'Toggle')
        else:
            raise ValueError(
                f'"{value}" not a valid option. Please use On/Off/Toggle/True/False/0/1/-1')

    def disable(self):
        """
        Distable the window
        
        :return: 
        """
        self.set('Disable', '')

    def enable(self):
        """
        Enable the window
        
        :return: 
        """
        self.set('Enable', '')

    def redraw(self):
        self.set('Redraw', '')

    def to_bottom(self):
        """
        Send window to bottom (behind other windows)
        :return:
        """
        self.set('Bottom', '')

    def to_top(self):
        """
        Bring the window to the foreground (above other windows)
        
        :return: 
        """
        self.set('Top', '')

    def _render_template(self, *args, **kwargs):
        kwargs['win'] = self
        return self.engine.render_template(*args, **kwargs)

    def _base_method(self, command, seconds_to_wait="", blocking=False):
        script = self._render_template(
            "window/base_command.ahk",
            command=command,
            title=f"ahk_id {self.id}",
            seconds_to_wait=seconds_to_wait
        )

        return self.engine.run_script(script, blocking=blocking)

    def activate(self):
        """
        Activate the window.

        See also: `WinActivate`_

        .. _WinActivate: https://www.autohotkey.com/docs/commands/WinActivate.htm

        :return:
        """
        self._base_method("WinActivate")

    def activate_bottom(self):
        """
        Calls `WinActivateBottom`_ on the window

        .. _WinActivateBottom: https://www.autohotkey.com/docs/commands/WinActivateBottom.htm

        :return:
        """
        self._base_method("WinActivateBottom")

    def close(self, seconds_to_wait=""):
        """
        Closes the Window. See also: `WinClose`_

        .. _WinClose: https://www.autohotkey.com/docs/commands/WinClose.htm

        :param seconds_to_wait:
        :return:
        """
        self._base_method("WinClose", seconds_to_wait=seconds_to_wait)

    def hide(self):
        """
        Hides the window. See also: `WinHide`_

        .. _WinHide: https://www.autohotkey.com/docs/commands/WinHide.htm
        
        
        :return:
        """
        self._base_method("WinHide")

    def kill(self, seconds_to_wait=""):
        self._base_method("WinKill", seconds_to_wait=seconds_to_wait)

    def maximize(self):
        """
        maximize the window
        
        :return: 
        """
        self._base_method("WinMaximize")

    def minimize(self):
        """
        minimize the window
        
        :return: 
        """
        self._base_method("WinMinimize")

    def restore(self):
        """
        restore the window
        
        :return: 
        """
        self._base_method("WinRestore")

    def show(self):
        """
        show the window
        
        :return: 
        """
        self._base_method("WinShow")

    def wait(self, seconds_to_wait=""):
        """
        
        :param seconds_to_wait: 
        :return: 
        """
        self._base_method("WinWait", seconds_to_wait=seconds_to_wait, blocking=True)

    def wait_active(self, seconds_to_wait=""):
        """
        
        :param seconds_to_wait: 
        :return: 
        """
        self._base_method("WinWaitActive", seconds_to_wait=seconds_to_wait, blocking=True)

    def wait_not_active(self, seconds_to_wait=""):
        """
        
        :param seconds_to_wait: 
        :return: 
        """
        self._base_method("WinWaitNotActive", seconds_to_wait=seconds_to_wait, blocking=True)

    def wait_close(self, seconds_to_wait=""):
        """
        
        :param seconds_to_wait: 
        :return: 
        """
        self._base_method("WinWaitClose", seconds_to_wait=seconds_to_wait, blocking=True)

    def move(self, x='', y='', width=None, height=None):
        """
        Move the window to a position and/or change its geometry

        :param x: 
        :param y: 
        :param width: 
        :param height: 
        :return: 
        """
        script = self._render_template(
            'window/win_move.ahk',
            title=f"ahk_id {self.id}",
            x=x, y=y, width=width, height=height
        )
        self.engine.run_script(script)

    def send(self, keys, delay=10, raw=False, blocking=False, escape=False, press_duration=-1):
        """
        Send keystrokes directly to the window.
        Uses ControlSend
        https://autohotkey.com/docs/commands/Send.htm
        """
        if escape:
            keys = escape_sequence_replace(keys)
        script = self._render_template(
            'window/win_send.ahk',
            title=f"ahk_id {self.id}",
            keys=keys, raw=raw, delay=delay,
            press_duration=press_duration, blocking=blocking
        )
        return self.engine.run_script(script, blocking=blocking)

    def click(self, x, y, blocking=False):
        """
        Click at an x/y location on the screen.
        Uses ControlClick
        https://autohotkey.com/docs/commands/ControlClick.htm
        """
        script = self._render_template(
            'window/win_click.ahk',
            x=x, y=y, hwnd=f"ahk_id {self.id}"
        )
        return self.engine.run_script(script, blocking=blocking)

    def __eq__(self, other):
        if not isinstance(other, Window):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(repr(self))


class WindowMixin(ScriptEngine):
    def __init__(self, *args, **kwargs):
        self.window_encoding = kwargs.pop('window_encoding', None)
        super().__init__(*args, **kwargs)

    def win_get(self, title='', text='', exclude_title='', exclude_text='', encoding=None):
        encoding = encoding or self.window_encoding
        script = self.render_template(
            'window/get.ahk',
            subcommand='ID',
            title=title,
            text=text,
            exclude_text=exclude_text,
            exclude_title=exclude_title
        )
        ahk_id = self.run_script(script)
        return Window(engine=self, ahk_id=ahk_id, encoding=encoding)

    def win_set(self, subcommand, *args, blocking=True):
        script = self.render_template(
            'window/set.ahk',  subcommand=subcommand, *args, blocking=blocking)
        self.run_script(script, blocking=blocking)

    @property
    def active_window(self):
        return self.win_get(title='A')

    def _all_window_ids(self):
        script = self.render_template('window/id_list.ahk')
        result = self.run_script(script)
        return result.split('\n')[:-1]  # last one is always an empty string

    def windows(self):
        """
        Returns a list of windows

        :return:
        """
        windowze = []
        for ahk_id in self._all_window_ids():
            win = Window(engine=self, ahk_id=ahk_id, encoding=self.window_encoding)
            windowze.append(win)
        return windowze

    def find_windows(self, func=None, **kwargs):
        """
        Find all matching windows
        
        :param func: a callable to filter windows
        :param bool exact: if False (the default) partial matches are found. If True, only exact matches are returned
        :param kwargs: keywords of attributes of the window (has no effect if ``func`` is provided)
        
        :return: a generator containing any matching :py:class:`~ahk.window.Window` objects. 
        """
        if func is None:
            exact = kwargs.pop('exact', False)

            def func(win):
                for attr, expected in kwargs.items():
                    if exact:
                        result = getattr(win, attr) == expected
                    else:
                        result = expected in getattr(win, attr)
                    if result is False:
                        return False
                return True
        for window in filter(func, self.windows()):
            yield window

    def find_window(self, func=None, **kwargs):
        """
        Like ``find_windows`` but only returns the first found window
        
        
        :param func: 
        :param kwargs: 
        :return: a :py:class:`~ahk.window.Window` object or ``None`` if no matching window is found
        """
        with suppress(StopIteration):
            return next(self.find_windows(func=func, **kwargs))

    def find_windows_by_title(self, title, exact=False):
        """
        Equivalent to ``find_windows(title=title)```
        
        Note that ``title`` is a ``bytes`` object
        
        :param bytes title: 
        :param exact: 
        :return: 
        """
        for window in self.find_windows(title=title, exact=exact):
            yield window

    def find_window_by_title(self, *args, **kwargs):
        """
        Like ``find_windows_by_title`` but only returns the first result.
        
        :return: a :py:class:`~ahk.window.Window` object or ``None`` if no matching window is found
        """
        with suppress(StopIteration):
            return next(self.find_windows_by_title(*args, **kwargs))

    def find_windows_by_text(self, text, exact=False):
        """
        
        :param text: 
        :param exact: 
        :return: a generator containing any matching :py:class:`~ahk.window.Window` objects. 
        """
        for window in self.find_windows(text=text, exact=exact):
            yield window

    def find_window_by_text(self, *args, **kwargs):
        """
        
        :param args: 
        :param kwargs: 
        :return: a :py:class:`~ahk.window.Window` object or ``None`` if no matching window is found 
        """
        with suppress(StopIteration):
            return next(self.find_windows_by_text(*args, **kwargs))

    def find_windows_by_class(self, class_name, exact=False):
        """
        
        :param class_name: 
        :param exact: 
        :return: a generator containing any matching :py:class:`~ahk.window.Window` objects. 
        """
        for window in self.find_windows(class_name=class_name, exact=exact):
            yield window

    def find_window_by_class(self, *args, **kwargs):
        """
        
        :param args: 
        :param kwargs: 
        :return: a :py:class:`~ahk.window.Window` object or ``None`` if no matching window is found 
        """
        with suppress(StopIteration):
            return next(self.find_windows_by_class(*args, **kwargs))



import ast
from ahk.script import ScriptEngine
from typing import Tuple, Union, Optional


class ScreenMixin(ScriptEngine):
    def image_search(self, image_path: str,
                     upper_bound: Tuple[int, int]=(0, 0), lower_bound: Tuple[int, int]=None,
                     color_variation: int=None, coord_mode: str='Screen',
                     scale_height: int=None, scale_width: int=None,
                     transparent: str=None, icon: int=None) -> Union[Tuple[int, int], None]:
        """
        `AutoHotkey ImageSearch reference`_

        .. _AutoHotkey ImageSearch reference: https://autohotkey.com/docs/commands/ImageSearch.htm


        :param image_path: path to the image file e.g. C:\location\of\cats.png
        :param upper_bound: a two-tuple of X,Y coordinates for the upper-left corner of the search area e.g. (200, 400)
            defaults to (0,0)

        :param lower_bound: like ``upper_bound`` but for the lower-righthand corner of the search area e.g. (400, 800)
            defaults to screen width and height (lower right-hand corner; ``%A_ScreenWidth%``, ``%A_ScreenHeight%``).

        :param color_variation: Shades of variation (up or down) for the intensity of RGB for each pixel. Equivalent of
            ``*n`` option. Defaults to 0.
        
        :param coord_mode: the Pixel CoordMode to use. Default is 'Screen'
        :param scale_height: Scale height in pixels. Equivalent of ``*hn`` option
        :param scale_width: Scale width in pixels. Equivalent of ``*wn`` option
        :param transparent: Specific color in the image that will be ignored during the search. Pixels with the exact
            color given will match any color. Can be used with color names e.g. (Black, Purple, Yellow) found
            at https://https://www.autohotkey.com/docs/commands/Progress.htm#colors or hexadecimal values e.g.
            (0xFFFFAA, 05FA15, 632511). Equivalent of ``*TransN`` option

        :param icon: Number of the icon group to use. Equivalent of ``*Icon`` option

        :return: coordinates of the upper-left pixel of where the image was found on the screen; ``None`` if the image
            was not found

        :rtype: Union[Tuple[int, int], None]

        Note: when only scale_height or only scale_width are provided, aspect ratio is maintained by default.
        """

        if scale_height and not scale_width:
            scale_width = -1
        elif scale_width and not scale_height:
            scale_height = -1

        options = []
        if icon:
            options.append(f'Icon{icon}')
        if color_variation:
            options.append(color_variation)
        if transparent:
            options.append(f'Trans{transparent}')
        if scale_width:
            options.append(f'w{scale_width}')
            options.append(f'h{scale_height}')

        x1, y1 = upper_bound
        if lower_bound:
            x2, y2 = lower_bound
        else:
            x2, y2 = ('%A_ScreenWidth%', '%A_ScreenHeight%')
        script = self.render_template('screen/image_search.ahk',
                                      x1=x1, x2=x2, y1=y1, y2=y2,
                                      coord_mode=coord_mode,
                                      image_path=image_path,
                                      options=options)
        resp = self.run_script(script)
        try:
            return ast.literal_eval(resp)
        except SyntaxError:
            return None

    def pixel_get_color(self, x: int, y: int, coord_mode: str='Screen',
                        alt: bool=False, slow: bool=False, rgb=True) -> Union[str, None]:
        """
        `AutoHotkey PixelGetColor reference`_

        .. _AutoHotkey PixelGetColor reference: https://autohotkey.com/docs/commands/PixelGetColor.htm

        :param x: x coordinate
        :param y: y coordinate
        :param coord_mode:
        :param alt:
        :param slow:
        :param rgb: returns
        :return: the color as an RGB hexidecimal string;
        :rtype: str
        """

        options = []
        if slow:
            options.append('Slow')
        elif alt:
            options.append('Alt')
        if rgb:
            options.append('RGB')
        script = self.render_template('screen/pixel_get_color.ahk',
                                      x=x, y=y,
                                      coord_mode=coord_mode,
                                      options=options)
        resp = self.run_script(script)
        return resp

    def pixel_search(self, color: Union[str, int], variation: int=0,
                     upper_bound: Tuple[int, int]=(0, 0), lower_bound: Tuple[int, int]=None,
                     coord_mode: str='Screen', fast: bool=True, rgb: bool=True) -> Union[Tuple[int, int], None]:
        """
        `AutoHotkey PixelSearch reference`_

        .. _AutoHotkey PixelSearch reference: https://autohotkey.com/docs/commands/PixelSearch.htm

        :param color:
        :param variation:
        :param upper_bound:
        :param lower_bound:
        :param coord_mode:
        :param fast:
        :param rgb:
        :return: the coordinates of the pixel; None if the pixel is not found
        """
        options = []
        if fast:
            options.append('Fast')
        if rgb:
            options.append('RGB')
        x1, y1 = upper_bound
        if lower_bound:
            x2, y2 = lower_bound
        else:
            x2, y2 = ('%A_ScreenWidth%', '%A_ScreenHeight%')

        script = self.render_template('screen/pixel_search.ahk',
                                      x1=x1, y1=y1, x2=x2, y2=y2,
                                      coord_mode=coord_mode,
                                      color=color,
                                      variation=variation,
                                      options=options)
        resp = self.run_script(script)
        try:
            return ast.literal_eval(resp)
        except SyntaxError:
            return None



"""
Contains directive classes
"""

from types import SimpleNamespace


class DirectiveMeta(type):
    """
    Overrides __str__ so directives with no arguments can be used without instantiation
    Overrides __hash__ to make objects 'unique' based upon a hash of the str representation
    """
    def __str__(cls):
        return f"#{cls.__name__}"

    def __hash__(self):
        return hash(str(self))

    def __eq__(cls, other):
        return str(cls) == other


class Directive(SimpleNamespace, metaclass=DirectiveMeta):
    """
    Simple directive class
    They are designed to be hashable and comparable with string equivalent of AHK directive.
    Directives that don't require arguments do not need to be instantiated.
    """
    def __init__(self, **kwargs):
        super().__init__(name=self.name, **kwargs)
        self._kwargs = kwargs

    @property
    def name(self):
        return self.__class__.__name__

    def __str__(self):
        if self._kwargs:
            arguments = ' '.join(str(value) for key, value in self._kwargs.items())
        else:
            arguments = ''
        return f"#{self.name} {arguments}".rstrip()

    def __eq__(self, other):
        return str(self) == other

    def __hash__(self):
        return hash(str(self))


class AllowSameLineComments(Directive):
    pass


class ClipboardTimeout(Directive):
    def __init__(self, milliseconds=0, **kwargs):
        kwargs['milliseconds'] = milliseconds
        super().__init__(**kwargs)


class ErrorStdOut(Directive):
    pass


class HotKeyInterval(ClipboardTimeout):
    pass


class HotKeyModifierTimeout(HotKeyInterval):
    pass


class Include(Directive):
    def __init__(self, include_name, **kwargs):
        kwargs['include_name'] = include_name
        super().__init__(**kwargs)


class IncludeAgain(Include):
    pass


class InputLevel(Directive):
    def __init__(self, level, **kwargs):
        kwargs['level'] = level
        super().__init__(**kwargs)


class InstallKeybdHook(Directive):
    pass


class InstallMouseHook(Directive):
    pass


class KeyHistory(Directive):
    def __init__(self, limit=40, **kwargs):
        kwargs['limit'] = limit
        super().__init__(**kwargs)


class MaxHotkeysPerInterval(Directive):
    def __init__(self, value, **kwargs):
        kwargs['value'] = value
        super().__init__(**kwargs)


class MaxMem(Directive):
    def __init__(self, megabytes: int, **kwargs):
        if megabytes < 1:
            raise ValueError('megabytes cannot be less than 1')
        if megabytes > 4095:
            raise ValueError('megabytes cannot exceed 4095')
        kwargs['megabytes'] = megabytes
        super().__init__(**kwargs)


class MaxThreads(Directive):
    def __init__(self):
        raise NotImplemented


class MaxThreadsBuffer(Directive):
    def __init__(self):
        raise NotImplemented


class MaxThreadsPerHotkey(Directive):
    def __init__(self):
        raise NotImplemented


class MenuMaskKey(Directive):
    def __init__(self):
        raise NotImplemented


class NoEnv(Directive):
    pass


class NoTrayIcon(Directive):
    pass


class Persistent(Directive):
    pass


class SingleInstance(Directive):
    pass


class UseHook(Directive):
    pass


class Warn(Directive):
    pass


class WinActivateForce(Directive):
    pass



from ahk.script import ScriptEngine


class SoundMixin(ScriptEngine):
    def sound_beep(self, frequency=523, duration=150):
        """
        REF: https://autohotkey.com/docs/commands/SoundBeep.htm

        :param frequency: number between 37 and 32767
        :param duration: how long in milliseconds to play the beep
        :return: None
        """

        script = self.render_template('sound/beep.ahk', frequency=frequency, duration=duration)
        self.run_script(script)

    def sound_play(self, filename, blocking=True):
        """
        REF: https://autohotkey.com/docs/commands/SoundPlay.htm

        :param filename:
        :param blocking:
        :param wait:
        :return:
        """

        script = self.render_template('sound/play.ahk', filename=filename, wait=1, blocking=blocking)
        self.run_script(script, blocking=blocking)

    def sound_get(self, device_number=1, component_type='MASTER', control_type='VOLUME'):
        """
        REF: https://autohotkey.com/docs/commands/SoundGet.htm


        :param device_number:
        :param component_type:
        :param control_type:
        :return:
        """

        script = self.render_template('sound/sound_get.ahk')
        return self.run_script(script)

    def get_volume(self, device_number=1):
        """
        REF: https://autohotkey.com/docs/commands/SoundGetWaveVolume.htm


        :param device_number:
        :return:
        """
        script = self.render_template('sound/get_volume.ahk', device_number=device_number)
        result = self.run_script(script)
        return result

    def sound_set(self, value, device_number=1, component_type='MASTER', control_type='VOLUME'):
        """
        REF: https://autohotkey.com/docs/commands/SoundSet.htm


        :param value:
        :param device_number:
        :param component_type:
        :param control_type:
        :return:
        """

        script = self.render_template('sound/sound_set.ahk', value=value,
                                      device_number=device_number,
                                      component_type=component_type,
                                      control_type=control_type)
        self.run_script(script)

    def set_volume(self, value, device_number=1):
        """
        REF: https://autohotkey.com/docs/commands/SoundSetWaveVolume.htm

        :param value: percent volume to set volume to
        :param device_number:
        :return:
        """

        script = self.render_template('sound/set_volume.ahk', value=value, device_number=device_number)
        self.run_script(script)
